#ifndef _SPARC_BB_H_
#define _SPARC_BB_H_

#include <inttypes.h> // uint32_t

#include "common/h/Vector.h"
#include "util/h/kdrvtypes.h"

#include "sparc_instr.h"
#include "fnCode.h"
#include "sparcTraits.h"

// fwd decls to avoid recursive #includes:
class basicblock_t;
class sparc_reg_set;
class monotone_bitwise_dataflow_fn;

class sparc_bb {
public:
   typedef fnCode fnCode_t;
   typedef uint16_t bbid_t; //same as in basicblock.h

   static pdvector<bool> backwardsSlice(kptr_t succ_bb_startAddr,
					regset_t &regsToSlice,
					bool includeControlDependencies,
					const fnCode_t &theFnCode,
					unsigned numInsnBytesToSkipAtEnd,
					unsigned bbNumInsnBytes,
					kptr_t bbStartAddr,
					kptr_t bbEndAddrPlus1) {
     // numInsnBytesToSkipAtEnd: the number of instruction bytes out of this basic
     // block, counting from the bottom of the basic block, that should be ignored
     // when doing the slice.  Of course 0 means skip nothing and consider the
     // whole basic block when slicing.
     assert(numInsnBytesToSkipAtEnd <= bbNumInsnBytes);

     sparc_reg_set &regs2slice = (sparc_reg_set&)regsToSlice;

     // In the future, we may allow slicing by memory locations, not just regs.

     // succ_bb is needed to determine whether the delay slot insn of 'this'
     // (if any) should be included.

     // If this bb ends in a DCTI with a delay slot insn that gets annulled if succ_bb
     // is next executed, then the delay slot insn should be removed from the slice (by
     // changing 'numInsnBytesToSkipAtEnd').
     const kptr_t lastInsnAddrInBlock = bbEndAddrPlus1 - 4;
     // getEndAddr() returns end of last insn...add 1 to get start of next insn...
     // subtract 4 to get back to beginning of last insn.

     // Since a basic block never crosses a chunk boundary, we can get a chunk now
     // and use its quick version of get1Insn() rather than function's get1Insn().
     fnCode_t::const_iterator chunk_iter = theFnCode.findChunk(bbStartAddr, false);
     assert(chunk_iter != theFnCode.end());
     const fnCode_t::codeChunk &theCodeChunk = *chunk_iter;
   
     instr_t *last_insn = theCodeChunk.get1Insn(lastInsnAddrInBlock);

     const unsigned bbnumInsns =
       theCodeChunk.calcNumInsnsWithinBasicBlock(bbStartAddr,
						 bbEndAddrPlus1);

     if (succ_bb_startAddr != -1U && bbnumInsns >= 2 &&
	 numInsnBytesToSkipAtEnd == 0) {
        instr_t *second_to_last_insn = theCodeChunk.get1Insn(lastInsnAddrInBlock - 4);
      
       sparc_instr::sparc_cfi cfi;
       second_to_last_insn->getControlFlowInfo(&cfi);

       if (cfi.delaySlot != instr_t::dsAlways &&
	   cfi.delaySlot != instr_t::dsNone && 
	   (cfi.fields.isBranch && cfi.fields.isConditional)) {
	 // There's a delay slot, and a chance that it's skipped
	 bool skipIt = cfi.delaySlot == instr_t::dsNever; // so far
	 if (!skipIt) {
	   assert(cfi.delaySlot == instr_t::dsWhenTaken);
	   kptr_t branchInsnAddr = bbStartAddr + 4*(bbnumInsns-2);
	   long offset_bytes = succ_bb_startAddr - branchInsnAddr;

	   assert(cfi.destination == sparc_instr::controlFlowInfo::pcRelative);
	   bool branchWasTaken = (offset_bytes==cfi.offset_nbytes);
	   if (!branchWasTaken)
	     // if it wasn't taken, then it should be a fall-through
	     assert(succ_bb_startAddr == branchInsnAddr +
		    cfi.delaySlot == instr_t::dsNone ? 4 : 8);

	   // cfi.delaySlot == dsWhenTaken.  Hence, the branch insn is annulled.
	   // Hence, the delay slot is annulled if NOT taken.
	   skipIt = !branchWasTaken;
	 }

	 if (skipIt)
	   numInsnBytesToSkipAtEnd += last_insn->getNumBytes();
       }
     }

     // initialize the result: no instructions are in the slice
     pdvector<bool> result(bbnumInsns, false);

     // add the PC register to regsToSlice if bb ends in a *conditional* branch (unless
     // it falls outside of the range of insns to slice thanks to numInsnBytesToSkipAtEnd)
     if (includeControlDependencies && numInsnBytesToSkipAtEnd == 0)
       factorInControlDependencyInRegsToSlice(regsToSlice, theCodeChunk, 
					      bbStartAddr, bbEndAddrPlus1);

     unsigned lcv = bbnumInsns - 1 - (numInsnBytesToSkipAtEnd / 4);

     unsigned byteoffset = bbNumInsnBytes - 4 - numInsnBytesToSkipAtEnd;

     while (true) {
       if (regsToSlice.isempty())
	 return result;

       assert(byteoffset < bbNumInsnBytes);
       instr_t *instr = theCodeChunk.get1Insn(bbStartAddr + byteoffset);

       sparc_instr::sparc_ru ru;
       instr->getRegisterUsage(&ru);

       *((sparc_reg_set*)ru.definitelyUsedBeforeWritten) -= 
	 sparc_reg_set::regsThatIgnoreWrites;
       *((sparc_reg_set*)ru.maybeUsedBeforeWritten) -= 
	 sparc_reg_set::regsThatIgnoreWrites;

       const sparc_reg_set regsToSlice_definitelyWritten = 
	 *((sparc_reg_set*)ru.definitelyWritten) & regs2slice;
       const sparc_reg_set regsToSlice_maybeWritten = 
	 *((sparc_reg_set*)ru.maybeWritten) & regs2slice;

       if (!regsToSlice_definitelyWritten.isempty()) {
	 // The insn contributes to the value of some reg(s) we're slicing; thus, the
	 // insn is part of the slice.
	 result[lcv] = true;

	 // The insn kills the set regsToSlice_definitelyWritten (probably just a
	 // single entry in the set, by the way).  Since they're dead, no previous insn
         // can affect them insofar as the slice concerns.  Our register interest now
	 // shifts to registers _used_ by this insn.

	 regs2slice -= regsToSlice_definitelyWritten; // dead

	 regs2slice += *((sparc_reg_set*)ru.definitelyUsedBeforeWritten); 
	   // g0 has been removed -- good
	 regs2slice += *((sparc_reg_set*)ru.maybeUsedBeforeWritten); 
	   // g0 has been removed -- good
       }
       if (!regsToSlice_maybeWritten.isempty()) {
	 // This instr *might* contribute to the value of some regs we care about.
	 // Thus, add it to the slice.
	 result[lcv] = true;

	 // The regs written to cannot be declared dead.
	 ;

	 // Live regs which contribute to the slice: care about 'em:
	 regs2slice += *((sparc_reg_set*)ru.definitelyUsedBeforeWritten); 
	   // g0 has been removed -- good
	 regs2slice += *((sparc_reg_set*)ru.maybeUsedBeforeWritten); 
	   // g0 has been removed -- good
       }

       if (lcv == 0 && byteoffset == 0)
         break;
       else if (lcv == 0)
         assert(false); // should be both or neither
       else if (byteoffset == 0)
         assert(false); // should be both or neither

       byteoffset -= instr->getNumBytes();
       --lcv;
     }

     return result;
   }

   static void
   factorInControlDependencyInRegsToSlice(regset_t &regsToSlice,
                                          const fnCode_t::codeChunk &theCodeChunk,
					  kptr_t bbStartAddr,
					  kptr_t bbEndAddrPlus1) {
     // This routine adds the PC register to regsToSlice if the bb ends with a
     // *conditional* branch.

     const unsigned bb_num_insns =
       theCodeChunk.calcNumInsnsWithinBasicBlock(bbStartAddr,
						 bbEndAddrPlus1);
     if (bb_num_insns == 0)
       return;

   // the following doesn't leave room for an architecture whose conditional branches
   // only "sometimes" have a delay slot (depending on some bit in the instruction,
   // presumably).  I don't know of any (the annul bit of sparc notwithstanding:
   // whether set or not, a truly *conditional* branch [i.e. not always or never]
   // always has delay slot...whether it gets executed or not is another question
   // of course, but of no concern here)
   
     unsigned ndxOfPossibleBranchInsn = bb_num_insns - 1;
     if (ndxOfPossibleBranchInsn-- == 0)
       return;

     kptr_t addr = bbStartAddr;
     while (ndxOfPossibleBranchInsn--) {
       const sparc_instr *i = (const sparc_instr *)theCodeChunk.get1Insn(addr);
       addr += i->getNumBytes();
     }
   
     instr_t *possible_branch_insn = theCodeChunk.get1Insn(addr);

     sparc_instr::sparc_cfi cfi;
     possible_branch_insn->getControlFlowInfo(&cfi);

     if (!cfi.fields.controlFlowInstruction)
       return;

     if ((cfi.sparc_fields.isBicc || cfi.sparc_fields.isBPcc) &&
	 cfi.condition != sparc_instr::iccAlways &&
	 cfi.condition != sparc_instr::iccNever) {
       (sparc_reg_set&)regsToSlice += sparc_reg::reg_pc;
     }
     else if ((cfi.sparc_fields.isFBfcc || cfi.sparc_fields.isFBPfcc) &&
	      cfi.condition != sparc_instr::fccAlways &&
	      cfi.condition != sparc_instr::fccNever) {
       (sparc_reg_set&)regsToSlice += sparc_reg::reg_pc;
     }
     else if (cfi.sparc_fields.isBPr) {
       (sparc_reg_set&)regsToSlice += sparc_reg::reg_pc;
     }
   }

   // Sun's CC says templated methods can't be static.  I disagree.
   template <class calcEffectOfCall>
   static monotone_bitwise_dataflow_fn*
   doLiveRegAnalysis(kptr_t addrWhereToStart, // such as just past the end of the bb
                     kptr_t addrWhereToStop, // such as the startAddr of the bb
                     const fnCode_t &theFnCode,
                     const calcEffectOfCall &callOracle,
                     unsigned bbNumInsnBytes,
		     kptr_t bbStartAddr,
		     kptr_t bbEndAddrPlus1,
		     unsigned bbNumSuccs,
		     bbid_t bbSuccOf0);
};

#endif /* SPARC_BB_H_ */
